import java.awt.Color;
import java.awt.Point;

import javax.media.opengl.*;
import javax.media.opengl.glu.GLU;

/**
 * Draws an animated jelly fish made of basic geometric geometric shapes.
 * @author Damian Johnson
 * @version October 3 2007
 */
class MrJellyRenderer implements GLEventListener {
  private static final Color BACKGROUND = Color.BLACK;
  private static final int RATE = 50; // Speed of motions (higher = slower)
  private static final double MAX_BODY_ROTATION = 30; // Max angle body tilts in degrees
  private static final boolean ROBOT_CROSSEYED = false; // If true, eyes point opposite directions
  private static final int TENTACLE_SEGMENTS = 3; // Number of tentacle segments
  
  private static final Color ROBOT_BODY_COLOR = Color.BLUE;
  private static final Color ROBOT_TENTACLE_COLOR = Color.RED;
  private static final Color ROBOT_EYE_COLOR = Color.WHITE;
  private static final Color ROBOT_IRIS_COLOR = Color.BLACK;
  private static final GLU glu = new GLU();
  private int counter = 0; // Number of times the display method has been called
  
  /*
   * Distance from center to display's corner. This is used to calculate robot dimensions (allowing
   * dynamic resizing).
   */
  private int quadrantWidth = 0;
  private int quadrantHeight = 0;
  
  public void init(GLAutoDrawable drawable) {
    final GL gl = drawable.getGL();
    
    // Sets the background
    float bgR = BACKGROUND.getRed() / 255;
    float bgG = BACKGROUND.getGreen() / 255;
    float bgB = BACKGROUND.getBlue() / 255;
    gl.glClearColor(bgR, bgG, bgB, 1.0f);
  }
  
  public void display(GLAutoDrawable drawable) {
    ++this.counter;
    
    final GL gl = drawable.getGL();
    gl.glClear(GL.GL_COLOR_BUFFER_BIT); // Clears the screen
    
    gl.glMatrixMode(GL.GL_MODELVIEW);
    gl.glLoadIdentity();
    
    double bodyRotation =
        MAX_BODY_ROTATION * 2 * Math.abs((float) this.counter / RATE % 2 - 1.0) - MAX_BODY_ROTATION;
    gl.glRotatef((float) bodyRotation, 0f, 0f, 1f);
    
    // Draws the tentacles
    setGlColor(gl, ROBOT_TENTACLE_COLOR);
    gl.glPushMatrix();
    int tentacleLength = (int) (this.quadrantHeight * 0.8 / TENTACLE_SEGMENTS);
    gl.glTranslatef((float) (this.quadrantWidth * -0.62), (float) (-tentacleLength / 2), 0f);
    float angle = (float) (50 * Math.abs(((double) this.counter / RATE / 2.3) % 2 - 1.0)) - 25;
    
    for (int i = 0; i < 6; ++i) {
      int angleMultiplier = 0; // Multiplier based on position
      if (i < 3) angleMultiplier = i - 3;
      else angleMultiplier = i - 2;
      
      drawTentacle(gl, angleMultiplier * angle);
      gl.glTranslatef((float) (this.quadrantWidth * 0.25), 0f, 0f); // Offset for next tentacle
    }
    gl.glPopMatrix();
    
    // Draws the body
    setGlColor(gl, ROBOT_BODY_COLOR);
    drawHead(gl);
    
    // Draws the eyes
    gl.glPushMatrix();
    float eyeX = (float) -(this.quadrantWidth * 0.2);
    float eyeY = (float) (this.quadrantHeight * 0.4);
    gl.glTranslatef(eyeX, eyeY, 0f);
    
    double xIris = Math.abs((double) this.counter / RATE * 1.6 % 2 - 1.0);
    double yIris = Math.abs(((double) this.counter / RATE / 2.3) % 2 - 1.0);
    drawEye(gl, xIris, yIris); // Left eye
    
    if (ROBOT_CROSSEYED) {
      xIris = Math.abs(xIris - 1);
      yIris = Math.abs(yIris - 1);
    }
    
    gl.glTranslatef(eyeX * -2, 0f, 0f);
    drawEye(gl, xIris, yIris); // Right eye
    gl.glPopMatrix();
    
    gl.glFlush(); // The double buffering was handled by initialization capacities
  }
  
  public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {}
  
  // Generates an orthogonal projection centered at (0, 0)
  public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
    final GL gl = drawable.getGL();
    gl.glViewport(0, 0, width, height);
    gl.glMatrixMode(GL.GL_PROJECTION);
    gl.glLoadIdentity();
    glu.gluOrtho2D(-width, width, -height, height);
    
    this.quadrantWidth = width;
    this.quadrantHeight = height;
  }
  
  // Draws trapezoid for the head
  private void drawHead(GL gl) {
    Point ur = new Point(); // Upper right point of head
    ur.x = (int) (this.quadrantWidth * 0.4);
    ur.y = (int) (this.quadrantHeight * 0.7);
    
    Point lr = new Point(); // Lower right point of head
    lr.x = (int) (this.quadrantWidth * 0.7);
    lr.y = (int) (this.quadrantHeight * 0.1);
    
    gl.glBegin(GL.GL_QUADS);
    gl.glVertex2i(ur.x, ur.y);
    gl.glVertex2i(lr.x, lr.y);
    gl.glVertex2i(-lr.x, lr.y);
    gl.glVertex2i(-ur.x, ur.y);
    gl.glEnd();
  }
  
  // Draws an eye with the iris at the current center
  // Iris offsets determine the percentage distance from lower left
  // Note: This overwrites the GL color
  private void drawEye(GL gl, double xIrisOffset, double yIrisOffset) {
    if (xIrisOffset > 1 || xIrisOffset < 0 || yIrisOffset > 1 || yIrisOffset < 0) {
      throw new IllegalArgumentException("Offsets out of bounds: " + xIrisOffset + "/"
          + yIrisOffset);
    }
    
    // Draws eye
    int xEye = (int) (this.quadrantWidth * 0.1);
    int yEye = (int) (this.quadrantHeight * 0.1);
    setGlColor(gl, ROBOT_EYE_COLOR);
    
    gl.glBegin(GL.GL_QUADS);
    gl.glVertex2i(xEye, yEye);
    gl.glVertex2i(xEye, -yEye);
    gl.glVertex2i(-xEye, -yEye);
    gl.glVertex2i(-xEye, yEye);
    gl.glEnd();
    
    // Draws iris
    int xIris = (int) (xEye * xIrisOffset); // Right coordinate
    int yIris = (int) (yEye * yIrisOffset); // Right coordinate
    int irisWidth = (int) (this.quadrantWidth * 0.1);
    int irisHeight = (int) (this.quadrantHeight * 0.1);
    setGlColor(gl, ROBOT_IRIS_COLOR);
    
    gl.glBegin(GL.GL_QUADS);
    gl.glVertex2i(xIris, yIris);
    gl.glVertex2i(xIris, yIris - irisHeight);
    gl.glVertex2i(xIris - irisWidth, yIris - irisHeight);
    gl.glVertex2i(xIris - irisWidth, yIris);
    gl.glEnd();
  }
  
  // Draws a tentacle with its top at the current center
  private void drawTentacle(GL gl, float segmentRotation) {
    int tentacleLength = (int) (this.quadrantHeight * 0.8 / TENTACLE_SEGMENTS);
    int tentacleWidth = (int) (this.quadrantWidth * 0.05);
    
    for (int i = 0; i < TENTACLE_SEGMENTS; ++i) {
      gl.glPushMatrix();
      for (int k = 0; k < i; ++k) {
        int xOffset =
            (int) (Math.cos(Math.toRadians(90 - segmentRotation * (k + 1))) * tentacleLength);
        int yOffset =
            (int) (Math.sin(Math.toRadians(90 + segmentRotation * (k + 1))) * tentacleLength);
        gl.glTranslatef(xOffset, -yOffset, 0f);
      }
      
      // Rotates about the upper left corner if rotation is positive, upper right if negative
      int widthTranslation =
          segmentRotation > 0 ? -tentacleWidth / 2 : tentacleWidth / 2;
      gl.glTranslatef(widthTranslation, tentacleLength, 0f);
      gl.glRotatef((i + 1) * segmentRotation, 0f, 0f, 1f);
      gl.glTranslatef(-widthTranslation, -tentacleLength, 0f);
      
      if (i != TENTACLE_SEGMENTS - 1) {
        // Central segment
        gl.glBegin(GL.GL_QUADS);
        gl.glVertex2i(-tentacleWidth, 0);
        gl.glVertex2i(tentacleWidth, 0);
        gl.glVertex2i(tentacleWidth, tentacleLength);
        gl.glVertex2i(-tentacleWidth, tentacleLength);
        gl.glEnd();
      } else {
        // Last segment
        gl.glBegin(GL.GL_TRIANGLES);
        gl.glVertex2i(0, 0);
        gl.glVertex2i(tentacleWidth, tentacleLength);
        gl.glVertex2i(-tentacleWidth, tentacleLength);
        gl.glEnd();
      }
      
      gl.glPopMatrix();
    }
  }
  
  private void setGlColor(GL gl, Color color) {
    float r = color.getRed() / 255;
    float g = color.getGreen() / 255;
    float b = color.getBlue() / 255;
    gl.glColor3f(r, g, b);
  }
}
